class Trace(object):
    _instance = None

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super(Trace, cls).__new__(cls, *args, **kwargs)
        return cls._instance

    def setup(self, verb=None):
        # Initialize and set up the trace instance
        self.verbose = verb if verb is not None else self.verbose if hasattr(self, "verbose") else True
        self.trace = print if self.verbose else lambda *a, **k: None
        return self.trace


def enable_trace(f):
    def wrapper(*args, **kwargs):
        f.__globals__['trace'] = Trace().setup()
        try:
            return f(*args, **kwargs)
        except Exception as e:
            # Handle exceptions here, e.g., log or raise the exception
            raise e
    return wrapper



import argparse
from pathlib import Path
from .defines import Scope, PROC_STAGE


def process_arguments(args=None):
    """
    Process command-line arguments using argparse.

    Parameters:
    - args: A list of command-line arguments
            (default is None, which means using sys.argv).

    Returns:
    - A dictionary containing the parsed command-line arguments.
    """

    # Create an ArgumentParser object
    parser = argparse.ArgumentParser(
        argument_default=argparse.SUPPRESS,
        prog="MetadataETL",
        description="Extract metadata from research data \
            and load it into the metadata catalog.",
        epilog="European XFEL GmbH."
    )

    # Define command-line arguments
    parser.add_argument('-b', '--base_folder', help='Base folder (default: $PWD)')
    parser.add_argument('-p', '--proposal', help='Full proposal number')
    parser.add_argument('-r', '--run', help='Run numbers', type=int)
    parser.add_argument('files', help='List of data file(s)', nargs='*', default=[])
    # We can also use the option type=argparse.FileType('r')
    # to check that the provided file names/paths are valid
    parser.add_argument('-s', '--scope', default=Scope.PROPOSAL, choices=Scope.get_scopes())
    parser.add_argument('-d', '--data', help='Input file for data and metadata specifications')
    parser.add_argument('-g', '--stage', default="load", choices=PROC_STAGE)
    parser.add_argument('-v', '--verbose', help='Verbose mode', default=False, action='store_true')

    # Parse the command-line arguments and store them in a dictionary
    # args = vars(parser.parse_args(args))
    return parser.parse_args(args)



import json
import h5py
from pathlib import Path


def visitor_func(name, node):
    if isinstance(node, h5py.Dataset):
        print(f"{name} is a dataset")
    else:
        print(f"{name} is a group")


class H5DataIndex:
    def __init__(self, prefix="CONTROL"):
        # Store an empty list for dataset names
        self.names = []
        self.groups = []
        self.files = []
        self.prefix = prefix
        self.data = {prefix: {}}

    def __call__(self, name, h5obj):
        # only h5py datasets have dtype attribute, so we can search on this
        # if hasattr(h5obj,'dtype') and not name in self.names:
        if isinstance(h5obj, h5py.Dataset):
            self.names += [name]
            self.insert_path(name, {"file": len(self.files) - 1, "key": name})
        else:
            # node is a group
            # print(f"{name} is a group")
            self.groups += [name]
            pass

    def insert_path(self, name, metadata):
        parent = self.data
        key = self.prefix
        cursor = parent.get(key, {})
        for p in name.split('/')[:-1]:
            if p in {'MDL'}:
                continue
            # if p in {'value', 'timestamp'}:
                # parent[key] = 0
                # continue
            if p not in cursor:
                cursor[p] = {}

            parent = cursor
            key = p
            cursor = parent.get(key, {})
        parent[key] = metadata

    def get_metadata(self, name):
        parent = self.data
        key = self.prefix
        cursor = parent.get(key, {})
        for p in name.split('/')[1:]:
            if p in {'MDL'}:
                continue
            if p not in cursor:
                return {"error": f"Path does not exist, starting from {p}"}
            parent = cursor
            key = p
            cursor = parent.get(key)
        return cursor | {"file": str(self.files[cursor.get("file")])}

    def register_file(self, filename):
        self.files += {filename}


class H5JsonEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, H5DataView):
            return obj.to_json()
        elif isinstance(obj, H5FileView):
            return obj.to_json()
        return json.JSONEncoder.default(self, obj)


class H5DataView:

    def __init__(self, flat_layout=True) -> None:
        self.prefix = 'CONTROL'
        self.flat_layout = flat_layout
        self.data = {self.prefix: {}}
        pass

    def __call__(self, name, h5obj):
        if isinstance(h5obj, h5py.Dataset):
            self.insert_path(name, {"metadata": "meta data"})

    def __str__(self) -> str:
        return str(self.data)

    def to_json(self):
        return self.data

    def insert_path(self, name, metadata):
        name = self.sanitize_name(name)
        if self.flat_layout:
            self.data[self.prefix][name] = metadata
            return

        parent = self.data
        key = self.prefix
        cursor = parent.get(key, {})
        for p in name.split('/'):
            if p in {'MDL'}:
                continue
            if p not in cursor:
                cursor[p] = {}

            parent = cursor
            key = p
            cursor = parent.get(key, {})
        parent[key] = metadata

    def get_metadata(self, name):
        name = self.sanitize_name(name)
        if self.flat_layout:
            return self.data[self.prefix].get(name)

        parent = self.data
        key = self.prefix
        cursor = parent.get(key, {})
        for p in name.split('/'):
            if p in {'MDL'}:
                continue
            if p not in cursor:
                return None
                return {"error": f"Path does not exist, starting from {p}"}
            parent = cursor
            key = p
            cursor = parent.get(key)
        return cursor

    def sanitize_name(self, name):
        return name.removeprefix(self.prefix + '/').removesuffix('/value').removesuffix('/timestamp')


class H5FileView:
    def __init__(self, prefix="CONTROL"):
        self.prefix = prefix
        self.data = {}

    def __call__(self, filename):
        if h5py.is_hdf5(filename):
            data_view = self.register_file(filename)
            try:
                with h5py.File(filename, 'r') as f:
                    # this will visit all objects inside the hdf5 file and store datasets in data_view object (ie names)
                    f["CONTROL"].visititems(data_view)
            except Exception as e:
                logger.error(e)
        else:
            raise SystemError(f"Invalid HDF5 file: {filename}")

    def __str__(self) -> str:
        return str(self.data)

    def to_json(self):
        return self.data

    def register_file(self, filename):
        self.data[str(filename)] = H5DataView(flat_layout=True)
        return self.data.get(str(filename))



import os
from extract.DataLoader import DataLoader
import datadir


def test_load_data(datadir):
    assert set(os.listdir(str(datadir))) == {
        "data",
        "parameters.json",
        "dataset.data",
        "specs.txt",
    }
    with (datadir / "specs.txt").open() as fp:
        contents = fp.read()
    assert contents == "Hello, world!\n"

    base_dir = datadir / "data"
    filename = "RAW-R0190-DA03-S00000.h5"

    assert set(os.listdir(str(base_dir))) == {
        filename,
    }



import h5py


def clone_datasets_simple(source_file, target_file, datasets):
    with h5py.File(target_file, 'w') as f_trgt, h5py.File(source_file, 'r') as f_src:
        f_src.copy(f_src[f'METADATA'], f_trgt)
        for dataset in datasets:
            print(dataset)
            group_path = f_src[dataset].parent.name
            for member in f_src[group_path]:
                if isinstance(f_src[f'{group_path}/{member}'], h5py.Dataset):
                    f_src.copy(f_src[f'{group_path}/{member}'], f_trgt, f'{group_path}/{member}')


def clone_datasets_special(source_file, target_file, datasets):
    with h5py.File(target_file, 'w') as f_trgt, h5py.File(source_file, 'r') as f_src:
        f_src.copy(f_src[f'METADATA'], f_trgt)

        for dataset in datasets:
            # Get the name of the parent for the group we want to copy
            group_path = f_src[dataset].parent.name
            print(dataset)
            group_id = f_trgt.require_group(group_path)

            for member in f_src[group_path]:
                src_dataset = f_src[f'{group_path}/{member}']
                shape = src_dataset.shape
                dtype = src_dataset.dtype
                attrs = src_dataset.attrs

                print(attrs.keys())

                print(shape)
                ds_arr = f_src[f'{group_path}/{member}'][:]
                ds = f_trgt.create_dataset(f'{group_path}/{member}', data=f_src[f'{group_path}/{member}'][:10])

                attrs = ds.attrs
                attrs.update(src_dataset.attrs)
                print(attrs.keys())



from http import HTTPStatus

STAT_OPERATORS = ['count', 'mean', 'std', 'min', 'max', 'med']

PROC_STAGE = ['config', 'extract', 'transform', 'load']

MAX_RETRIES = 3
TIMEOUT = 12
SSL_VERIFY = False
PAGE_NUMBER = 1
PAGE_SIZE = 100

OK = HTTPStatus.OK
GET = 'GET'

'''
Scope: defines the scope of data processing
 - proposal: the processing spans on all runs and files,
             and the summary is aggregated at the specified proposal
 - run: the processing spans on all files for the specified run,
        and the summary is generated per run
 - file: The processing is done per file,
         and a summary is also generated for each individual file
'''


class Scope:
    # Constants representing different data processing scopes
    PROPOSAL = 'proposal'
    FILE = 'file'
    RUN = 'run'

    @staticmethod
    def get_scopes():
        """
        Return a list of all available data processing scopes.

        Returns:
        - A list containing strings representing different data processing scopes.
        """
        return [Scope.PROPOSAL, Scope.FILE, Scope.RUN]



from extract.DataLoader import DataLoader
from mdc_api import connect_to_service, retrieve_units, retrieve_unit_info, ServiceInterface, inject_parameter, retrieve_parameter, delete_parameter
import unittest
import requests
from http import HTTPStatus

app_info = {
    "id": "6TqGFOWg0PzEXsHCs9baMmGQ7Hc6UhpU7YztJz7Y9OI",
    "secret": "N1_qi7Zt7ElHhJksM5ocnkGsvHS0jnp03WOMS9Hpzz8",
    "url": "https://192.168.56.101:3000/dev_metadata",
    "scope": ""
}

user_info = {
    "name": "Djelloul Boukhelef",
    "email": "djelloul.boukhelef@xfel.eu",
    "provider": "ldap",
    "uid": "boukhele"
}


class ParameterApiTest(unittest.TestCase):
    def setUp(self) -> None:
        # need to setup mdc for testing
        return
        self.service_interface = connect_to_service(app_info, user_info)
        self.client = self.service_interface.oauth_client
        self.session = self.service_interface.oauth_session
        assert (self.service_interface is not None)
        assert (self.client is not None)
        assert (self.session is not None)

        res = ServiceInterface.get_access_token(app_info["url"] + "/oauth/token", self.service_interface.client_id, self.service_interface.client_secret)
        assert (res.get("expires_in", None))

    def test_retrieve_units(self):
        # need to setup mdc for testing
        return
        service_interface = self.service_interface

        units = retrieve_units(service_interface)
        unit = retrieve_unit_info(service_interface, 10)

        assert unit['id'] == 10
        assert unit['name'] == "Hertz"
        assert unit['identifier'] == "HERTZ"
        assert unit['symbol'] == "Hz"

        assert unit in units

        unit = retrieve_unit_info(service_interface, 400)
        assert unit["info"] == "Resource not found!"

    def test_create_parameter(self):
        # need to setup mdc for testing
        return
        service_interface = self.service_interface
        parameter = {
            'data_source': 'MID/XTD9/XGM/MAIN/CAR',
            'name': 'Liquid velocity',
            'value': 445.6,
            'minimum': 305.0,
            'maximum': 1064.0,
            'mean': 750.5,
            'standard_deviation': 7.5,
            'data_type_id': 20,
            'parameter_type_id': 1,
            'unit_id': 1,
            'unit_prefix': '',
            'flg_available': True,
            'description': 'Dummy parameter for testing purpose only!!',
            'data_groups_parameters_attributes': [{'data_group_id': 4}]
            #  'runs_parameters_attributes': [{'run_id': 4}]
        }

        #  Test inject new parameter
        res = inject_parameter(service_interface, parameter)
        assert "id" in res
        parameter_id = res['id']

        #  Test retrieve new parameter
        parameter_data = retrieve_parameter(service_interface, parameter_id)

        assert parameter_id == parameter_data["id"]

        #  Test delete parameter
        res = delete_parameter(service_interface, parameter_id)
        assert res.status_code == HTTPStatus.NO_CONTENT

        #  Make sure that the parameter no longer exists after the deletion
        res = retrieve_parameter(service_interface, parameter_id)
        assert "id" not in res
        assert "info" in res



import unittest
from unittest.mock import mock_open, patch
from utils.utils import flatten_list, load_data_from_file, store_data_to_file, create_parameter_data, create_metadata_client

import json
from loguru import logger


class TestFlattenList(unittest.TestCase):
    def test_flat_list(self):
        self.assertEqual(flatten_list([1, 2, 3]), [1, 2, 3])

    def test_nested_list(self):
        self.assertEqual(flatten_list([1, [2, 3], 4]), [1, 2, 3, 4])

    def test_deeply_nested_list(self):
        self.assertEqual(flatten_list([1, [2, [3, [4, 5]], 6], 7]), [1, 2, 3, 4, 5, 6, 7])

    def test_empty_list(self):
        self.assertEqual(flatten_list([]), [])

    def test_list_with_empty_nested_lists(self):
        self.assertEqual(flatten_list([[], [1, [], 2], 3]), [1, 2, 3])

    def test_list_with_non_integer_elements(self):
        self.assertEqual(flatten_list(["a", ["b", ["c"]], "d"]), ["a", "b", "c", "d"])

    def test_list_with_mixed_data_types(self):
        self.assertEqual(flatten_list([1, ["a", [3.14, [True]]], None]), [1, "a", 3.14, True, None])

    def test_lists_with_concatenations(self):

        l1 = [1, 2, 3, 4]
        l2 = ['a', l1, 'b']
        l3 = ['x', l2, 'y', l1, 'z']
        l4 = [[l1, [l2, [l3]]], [[[l1], l2], l3]]

        fl1 = flatten_list(l1)
        fl2 = flatten_list(l2)
        fl3 = flatten_list(l3)
        fl4 = flatten_list(l4)

        self.assertEqual(fl1, l1)
        self.assertEqual(fl2, ['a'] + fl1 + ['b'])
        self.assertEqual(fl3, ['x'] + fl2 + ['y'] + fl1 + ['z'])
        self.assertEqual(fl4, fl1 + fl2 + fl3 + fl1 + fl2 + fl3)


class TestLoadStoreDataFromAndToFile(unittest.TestCase):
    @patch("builtins.open", new_callable=mock_open, read_data='{"key": "value"}')
    def test_valid_json_file(self, mock_file):
        self.assertEqual(load_data_from_file("dummy.json"), {"key": "value"})

    @patch("builtins.open", new_callable=mock_open, read_data='invalid json')
    @patch("loguru.logger.error")
    def test_invalid_json_file(self, mock_logger, mock_file):
        self.assertIsNone(load_data_from_file("dummy.json"))
        mock_logger.assert_called()

    @patch("builtins.open", side_effect=FileNotFoundError("File not found"))
    @patch("loguru.logger.error")
    def test_file_not_found(self, mock_logger, mock_file):
        self.assertIsNone(load_data_from_file("nonexistent.json"))
        mock_logger.assert_called()

    @patch("builtins.open", new_callable=mock_open, read_data='{"name": "Dj", "age": 76, "height":12345}')
    def test_store_data_to_file(self, mock_file):
        data = {"name": "Dj", "age": 76, "height": 12345}
        store_data_to_file(data, "test.json")
        mock_file.assert_called_once_with("test.json", 'w', encoding='utf-8')
        handle = mock_file()
        handle.write.assert_called()
        self.assertEqual(load_data_from_file("test.json"), data)


class TestCreateParameterData(unittest.TestCase):
    @patch("loguru.logger.warning")
    @patch("loguru.logger.error")
    def test_create_parameter_data_valid(self, mock_error, mock_warning):
        key = "param1"
        data = {'med': 50, 'min': 40, 'max': 60, 'mean': 50, 'std': 5}
        attrs = {'device': 'DS/1', 'name': 'Temperature', 'unit_id': 1, 'unit_prefix_name': 'C', 'alias': 'Temp'}

        expected = {
            'data_source': 'DS/1', 'name': 'Temperature', 'value': 50, 'minimum': 40, 'maximum': 60,
            'mean': 50, 'standard_deviation': 5, 'data_type_id': 20, 'parameter_type_id': 1, 'unit_id': 1,
            'unit_prefix': 'C', 'flg_available': True, 'description': 'Temp'
        }

        result = create_parameter_data(key, data, attrs)

        self.assertEqual(result, expected)
        mock_warning.assert_not_called()  # No warning should be logged
        mock_error.assert_not_called()  # No error should be logged

    @patch("loguru.logger.warning")
    @patch("loguru.logger.error")
    def test_create_parameter_data_invalid_mean(self, mock_error, mock_warning):
        key = "param2"
        data = {'med': 45, 'min': 40, 'max': 60, 'mean': 70, 'std': 5}
        attrs = {'device': 'DS/2', 'name': 'Pressure', 'unit_id': 2, 'unit_prefix_name': 'Pa', 'alias': 'Pres'}

        expected = {
            'data_source': 'DS/2', 'name': 'Pressure', 'value': 45, 'minimum': 40, 'maximum': 60,
            'mean': 45, 'standard_deviation': 5, 'data_type_id': 20, 'parameter_type_id': 1, 'unit_id': 2,
            'unit_prefix': 'Pa', 'flg_available': True, 'description': 'Pres'
        }

        result = create_parameter_data(key, data, attrs)

        self.assertEqual(result, expected)
        mock_warning.assert_called()  # Warning should be logged
        mock_error.assert_not_called()  # No error should be logged

    @patch("loguru.logger.warning")
    @patch("loguru.logger.error")
    def test_create_parameter_data_with_error(self, mock_error, mock_warning):
        key = "param3"
        data = {'med': 30, 'min': 20, 'max': 40, 'mean': 30, 'std': 2}
        attrs = {'device': 'DS/3', 'name': 'Humidity', 'unit_id': 3, 'unit_prefix_name': 'g/m³'}

        # Simulate an error scenario where 'device' attribute is missing
        attrs_missing = {'name': 'Humidity', 'unit_id': 3, 'unit_prefix_name': 'g/m³'}

        # This should raise an error due to missing 'device' in attrs
        result = create_parameter_data(key, data, attrs_missing)

        self.assertIsNone(result)  # The result should be None because of the error
        mock_warning.assert_not_called()  # No warning should be logged
        mock_error.assert_called()  # An error should be logged


if __name__ == "__main__":
    unittest.main()




proposal_info_response = {
    "id": 923,
    "number": 900390,
    "title": "FXE commissioning",
    "abstract": None,
    "proposal_folder": "p900390",
    "def_proposal_path": "/gpfs/exfel/exp/FXE/202331/p900390/raw",
    "doi": "10.22003/XFEL.EU-DATA-900390-00",
    "url": None,
    "instrument_id": 2,
    "instrument_cycle_id": 341,
    "principal_investigator_id": 658,
    "main_proposer_id": 180,
    "local_contact_id": 180,
    "local_data_contact_id": None,
    "experiment_data_contact_id": 180,
    "orig_repository_id": 103,
    "last_run": 704,
    "begin_at": "2023-07-01T00:02:00.000+02:00",
    "end_at": "2025-01-31T23:58:00.000+01:00",
    "release_at": "2028-12-31T23:58:00.000+01:00",
    "num_delivered_shifts": None,
    "num_preparation_shifts": None,
    "flg_public": False,
    "flg_proposal_system": "L",
    "flg_preferred_data_output": "P",
    "flg_available": True,
    "flg_auto_run_quality": 1,
    "flg_beamtime_status": "R",
    "beamtime_start_at": None,
    "beamtime_end_at": None,
    "flg_cloneable_status": "SC",
    "flg_data_archived_notification": "R",
    "expected_data_archive_at": "2027-02-12T00:00:00.000+01:00",
    "data_archived_at": None,
    "description": None,
    "data_plan_comments": None,
    "flg_visibility_status": "",
    "logbook_info": {
    },
    "beamtimes": [
        {
            "id": 1064,
            "begin_at": "2023-07-01T00:02:00.000+02:00",
            "end_at": "2026-12-31T23:58:00.000+01:00",
            "flg_available": True,
            "description": None
        }
    ],
    "instrument_identifier": "FXE",
    "meeting": None,
    "team": [
        {
            "id": 19191,
            "user_id": 1234,
            "data_access_role": 1,
            "flg_lab_access": False,
            "flg_support": False
        },
    ],
    "users_ids": [
        [
            1234,
            "jbald",
            "ldap"
        ],
    ],
    "openly_released_at": None,
    "techniques": [],
    "users_info": [
        {
            "user_id": 1234,
            "data_access_role": 2,
            "flg_lab_access": False,
            "flg_support": False,
            "uid": "jbald",
            "provider": "ldap",
            "name": "John Bald",
            "email": "jonh.bald@example.com",
            "flg_upex_acceptances": True
        },
    ]
}


proposal_runs_response = {
    "proposal": {
        "id": 923,
        "number": 900390,
        "title": "FXE commissioning"
    },
    "runs": [
        {
            "id": 214026,
            "run_number": 704,
            "run_folder": "r0704",
            "run_alias": "",
            "experiment_id": 4462,
            "sample_id": 4962,
            "begin_at": "2024-01-31T13:30:27.000+01:00",
            "end_at": "2024-01-31T13:31:04.000+01:00",
            "first_train": 1919430548,
            "last_train": 1919430895,
            "flg_available": True,
            "flg_status": 2,
            "original_format": "",
            "system_msg": "",
            "description": "",
            "created_at": "2024-01-31T13:30:27.000+01:00",
            "updated_at": "2024-01-31T13:33:26.000+01:00",
            "flg_run_quality": 1,
            "size": 5779619840,
            "num_files": 34,
            "flg_cal_data_status": "E",
            "cal_pipeline_reply": "Succeeded: none; Failed: FXE_DET_LPD1M-1",
            "migration_request_at": "2024-01-31T13:31:38.000+01:00",
            "migration_begin_at": "2024-01-31T13:31:48.000+01:00",
            "migration_end_at": "2024-01-31T13:32:07.000+01:00",
            "cal_num_requests": 1,
            "cal_last_request_at": "2024-01-31T13:31:38.000+01:00",
            "cal_last_begin_at": "2024-01-31T13:32:11.000+01:00",
            "cal_last_end_at": "2024-01-31T13:33:26.000+01:00",
            "flg_auto_run_quality": None,
            "run_technique_snapshot": None,
            "cal_input_path": None,
            "cal_output_path": None,
            "repositories": {
                "XFEL_GPFS_OFFLINE_RAW_CC": {
                    "identifier": "XFEL_GPFS_OFFLINE_RAW_CC",
                    "name": "XFEL GPFS offline Raw data in DESY CC",
                    "mount_point": "/gpfs/exfel/d/raw",
                    "data_groups": 1,
                    "flg_available": False
                },
                "DESY_DCACHE_RAW_CC": {
                    "identifier": "DESY_DCACHE_RAW_CC",
                    "name": "dCache Raw Data in DESY CC",
                    "mount_point": "/pnfs/desy.de/exfel/archive/XFEL/raw",
                    "data_groups": 1,
                    "flg_available": True
                }
            }
        }
    ]
}


run_info_response = {
    "id": 214026,
    "run_number": 704,
    "run_folder": "r0704",
    "run_alias": "",
    "experiment_id": 4462,
    "sample_id": 4962,
    "begin_at": "2024-01-31T13:30:27.000+01:00",
    "end_at": "2024-01-31T13:31:04.000+01:00",
    "first_train": 1919430548,
    "last_train": 1919430895,
    "migration_request_at": "2024-01-31T13:31:38.000+01:00",
    "migration_begin_at": "2024-01-31T13:31:48.000+01:00",
    "migration_end_at": "2024-01-31T13:32:07.000+01:00",
    "flg_available": True,
    "flg_status": 2,
    "flg_run_quality": 1,
    "size": 5779619840,
    "num_files": 34,
    "flg_cal_data_status": "E",
    "cal_pipeline_reply": "Succeeded: none; Failed: FXE_DET_LPD1M-1",
    "cal_num_requests": 1,
    "cal_last_request_at": "2024-01-31T13:31:38.000+01:00",
    "cal_last_begin_at": "2024-01-31T13:32:11.000+01:00",
    "cal_last_end_at": "2024-01-31T13:33:26.000+01:00",
    "original_format": "",
    "system_msg": "",
    "description": "",
    "experiment": {
        "id": 4462,
        "name": "Test DAQ",
        "doi": "xfel.mdc.exp.4462",
        "abstract": None,
        "experiment_folder": "e0001",
        "experiment_type_id": 10,
        "proposal_id": 923,
        "first_prefix_path": "/gpfs/exfel/exp/FXE/202331/p900390/raw/e0001/",
        "flg_available": True,
        "publisher": "XFEL",
        "contributor": "XFEL",
        "description": None
    },
    "sample": {
        "id": 4962,
        "name": "No sample",
        "url": None,
        "sample_type_id": 7,
        "flg_proposal_system": None,
        "proposal_system_id": None,
        "flg_available": True,
        "description": None
    },
    "data_groups_repositories": [
        {
            "id": 213902,
            "name": "raw_e0001_n0489",
            "language": "en",
            "doi": "xfel.mdc.dg.213902",
            "format": "",
            "data_passport": "",
            "experiment": {
                "id": 4462,
                "name": "Test DAQ",
                "doi": "xfel.mdc.exp.4462"
            },
            "data_group_type": {
                "id": 1,
                "name": "Raw",
                "identifier": "RAW"
            },
            "repositories": [
                {
                    "id": 11,
                    "name": "XFEL GPFS offline Raw data in DESY CC",
                    "identifier": "XFEL_GPFS_OFFLINE_RAW_CC",
                    "mount_point": "/gpfs/exfel/d/raw",
                    "transfer_agent_identifier": "offline_raw",
                    "transfer_agent_server_url": None,
                    "flg_context": "Global",
                    "flg_available": True,
                    "rank": 11,
                    "data_group_type_id": 1,
                    "description": "European XFEL GPFS offline Raw Data instance in DESY Computer Center"
                },
                {
                    "id": 21,
                    "name": "dCache Raw Data in DESY CC",
                    "identifier": "DESY_DCACHE_RAW_CC",
                    "mount_point": "/pnfs/desy.de/exfel/archive/XFEL/raw",
                    "transfer_agent_identifier": "dcache_raw",
                    "transfer_agent_server_url": "",
                    "flg_context": "Global",
                    "flg_available": True,
                    "rank": 21,
                    "data_group_type_id": 1,
                    "description": "DESY dCACHE Raw Data instance in DESY Computer Center"
                }
            ]
        }
    ],
    "techniques": [],
    "raw_repositories": [
        {
            "id": 11,
            "name": "XFEL GPFS offline Raw data in DESY CC",
            "identifier": "XFEL_GPFS_OFFLINE_RAW_CC",
            "img_url": "https://in.xfel.eu/metadata/assets/repos_host_logo/desy_logo_rgb-911077061ae7089b93c512ee0a396e4139e0a4d9dff3cf23fcb4d2a7f44eaf3b.jpg",
            "transfer_agent_identifier": "offline_raw",
            "transfer_agent_server_url": None,
            "mount_point": "/gpfs/exfel/d/raw",
            "flg_available": True,
            "flg_context": "Global",
            "data_group_type_id": 1,
            "rank": 11,
            "description": "European XFEL GPFS offline Raw Data instance in DESY Computer Center"
        },
        {
            "id": 21,
            "name": "dCache Raw Data in DESY CC",
            "identifier": "DESY_DCACHE_RAW_CC",
            "img_url": "https://in.xfel.eu/metadata/assets/repos_host_logo/desy_logo_rgb-911077061ae7089b93c512ee0a396e4139e0a4d9dff3cf23fcb4d2a7f44eaf3b.jpg",
            "transfer_agent_identifier": "dcache_raw",
            "transfer_agent_server_url": "",
            "mount_point": "/pnfs/desy.de/exfel/archive/XFEL/raw",
            "flg_available": True,
            "flg_context": "Global",
            "data_group_type_id": 1,
            "rank": 21,
            "description": "DESY dCACHE Raw Data instance in DESY Computer Center"
        }
    ],
    "reports": [
        {
            "id": 41,
            "name": "FXE_DET_LPD1M-1_correct_900390_r0704_240131_123138_453943.pdf",
            "cal_report_path": "/gpfs/exfel/exp/FXE/202331/p900390/usr/Reports/r0704/",
            "cal_report_at": "2024-01-31T13:31:38.453+01:00",
            "run_id": 214026,
            "description": "FXE_DET_LPD1M-1 detector corrections (errors occurred)"
        }
    ]
}


get_all_units_reponse = [{'id': 1, 'name': 'Number', 'identifier': 'NUMBER', 'symbol': '№', 'origin': 'Other', 'flg_available': True, 'description': 'Values without a clearly defined unit'}, {'id': 2, 'name': 'Count', 'identifier': 'COUNT', 'symbol': '#', 'origin': 'Other', 'flg_available': True, 'description': 'Counters, iteration variable'}, {'id': 3, 'name': 'Meter', 'identifier': 'METER', 'symbol': 'm', 'origin': 'SI base unit', 'flg_available': True, 'description': 'Unit of length'}, {'id': 4, 'name': 'Gram', 'identifier': 'GRAM', 'symbol': 'g', 'origin': 'SI base unit Kg * 0.001', 'flg_available': True, 'description': 'Unit of mass'}, {'id': 5, 'name': 'Second', 'identifier': 'SECOND', 'symbol': 's', 'origin': 'SI base unit', 'flg_available': True, 'description': 'Unit of time'}, {'id': 6, 'name': 'Ampere', 'identifier': 'AMPERE', 'symbol': 'A', 'origin': 'SI base unit', 'flg_available': True, 'description': 'Unit of electric current'}, {'id': 7, 'name': 'Kelvin', 'identifier': 'KELVIN', 'symbol': 'K', 'origin': 'SI base unit', 'flg_available': True, 'description': 'Unit of thermodynamic temperature'}, {'id': 8, 'name': 'Mole', 'identifier': 'MOLE', 'symbol': 'mol', 'origin': 'SI base unit', 'flg_available': True, 'description': 'Unit of amount of substance'}, {'id': 9, 'name': 'Candela', 'identifier': 'CANDELA', 'symbol': 'cd', 'origin': 'SI base unit', 'flg_available': True, 'description': 'Unit of luminous intensity'}, {'id': 10, 'name': 'Hertz', 'identifier': 'HERTZ', 'symbol': 'Hz', 'origin': 'SI derived unit', 'flg_available': True, 'description': 'Unit of frequency (s-1)'}, {'id': 11, 'name': 'Radian', 'identifier': 'RADIAN', 'symbol': 'rad', 'origin': 'SI derived unit', 'flg_available': True, 'description': 'Unit of plane angle (2D angle)'}, {'id': 12, 'name': 'Degree', 'identifier': 'DEGREE', 'symbol': 'deg', 'origin': 'Non-SI unit mentioned in the SI', 'flg_available': True, 'description': 'Unit of plane angle (2D angle)'}, {'id': 13, 'name': 'Steradian', 'identifier': 'STERADIAN', 'symbol': 'sr', 'origin': 'SI derived unit', 'flg_available': True, 'description': 'Unit of solid angle (3D angle)'}, {'id': 14, 'name': 'Newton', 'identifier': 'NEWTON', 'symbol': 'N', 'origin': 'SI derived unit', 'flg_available': True, 'description': 'Mechanics unit of force, weight (kg·m·s-2)'}, {'id': 15, 'name': 'Pascal', 'identifier': 'PASCAL', 'symbol': 'Pa', 'origin': 'SI derived unit', 'flg_available': True, 'description': 'Mechanics unit of pressure, stress (kg·m-1·s-2)'}, {'id': 16, 'name': 'Joule', 'identifier': 'JOULE', 'symbol': 'J', 'origin': 'SI derived unit', 'flg_available': True, 'description': 'Mechanics unit of energy, work, heat (kg·m2·s-2)'}, {'id': 17, 'name': 'Electronvolt', 'identifier': 'ELECTRONVOLT', 'symbol': 'eV', 'origin': 'Non-SI unit mentioned in the SI', 'flg_available': True, 'description': 'Physics unit of energy'}, {'id': 18, 'name': 'Watt', 'identifier': 'WATT', 'symbol': 'W', 'origin': 'SI derived unit', 'flg_available': True, 'description': 'Mechanics unit of power (kg·m2·s-3)'}, {'id': 19, 'name': 'Coulomb', 'identifier': 'COULOMB', 'symbol': 'C', 'origin': 'SI derived unit', 'flg_available': True, 'description': 'Electromagnetism unit of charge (A·s)'}, {'id': 20, 'name': 'Volt', 'identifier': 'VOLT', 'symbol': 'V', 'origin': 'SI derived unit', 'flg_available': True, 'description': 'Electromagnetism unit of voltage (kg·m2·s-3·A-1)'}, {'id': 21, 'name': 'Farad', 'identifier': 'FARAD', 'symbol': 'F', 'origin': 'SI derived unit', 'flg_available': True, 'description': 'Electromagnetism unit of electric capacitance (kg-1·m-2·s4·A2)'}, {'id': 22, 'name': 'Ohm', 'identifier': 'OHM', 'symbol': 'Ω', 'origin': 'SI derived unit', 'flg_available': True, 'description': 'Electromagnetism unit of electric resistance (kg·m2·s-3·A-3)'}, {'id': 23, 'name': 'Siemens', 'identifier': 'SIEMENS', 'symbol': 'S', 'origin': 'SI derived unit', 'flg_available': True, 'description': 'Electromagnetism unit of electrical conductance (kg-1·m-2·s3·A2)'}, {'id': 24, 'name': 'Weber', 'identifier': 'WEBER', 'symbol': 'Wb', 'origin': 'SI derived unit', 'flg_available': True, 'description': 'Electromagnetism unit of magnetic flux (kg·m2·s-2·A-1)'}, {'id': 25, 'name': 'Tesla', 'identifier': 'TESLA', 'symbol': 'T', 'origin': 'SI derived unit', 'flg_available': True, 'description': 'Electromagnetism unit of magnetic Field (kg·s-2·A-1)'}, {'id': 26, 'name': 'Henry', 'identifier': 'HENRY', 'symbol': 'H', 'origin': 'SI derived unit', 'flg_available': True, 'description': 'Electromagnetism unit of inductance (kg·m2·s-2·A-2)'}, {'id': 27, 'name': 'Degree Celsius', 'identifier': 'DEGREE_CELSIUS', 'symbol': 'degC', 'origin': 'SI derived unit', 'flg_available': True, 'description': 'Unit of temperature'}, {'id': 28, 'name': 'Lumen', 'identifier': 'LUMEN', 'symbol': 'lm', 'origin': 'SI derived unit', 'flg_available': True, 'description': 'Optics unit of luminous flux (cd·sr)'}, {'id': 29, 'name': 'Lux', 'identifier': 'LUX', 'symbol': 'lx', 'origin': 'SI derived unit', 'flg_available': True, 'description': 'Optics unit of illuminance (cd·sr·m-2)'}, {'id': 30, 'name': 'Becquerel', 'identifier': 'BECQUEREL', 'symbol': 'Bq', 'origin': 'SI derived unit', 'flg_available': True, 'description': 'Radioactivity unit of radioactivity (s-1)'}, {'id': 31, 'name': 'Gray', 'identifier': 'GRAY', 'symbol': 'Gy', 'origin': 'SI derived unit', 'flg_available': True, 'description': 'Radioactivity unit of absorbed dose (m2·s-1)'}, {'id': 32, 'name': 'Sievert', 'identifier': 'SIEVERT', 'symbol': 'Sv', 'origin': 'SI derived unit', 'flg_available': True, 'description': 'Radioactivity unit of equivalent dose (m2·s-1)'}, {'id': 33, 'name': 'Katal', 'identifier': 'KATAL', 'symbol': 'kat', 'origin': 'SI derived unit', 'flg_available': True, 'description': 'Unit of catalytic activity (mol·s-1)'}, {'id': 34, 'name': 'Minute', 'identifier': 'MINUTE', 'symbol': 'min', 'origin': 'Non-SI unit mentioned in the SI', 'flg_available': True, 'description': 'Unit of time'}, {'id': 35, 'name': 'Hour', 'identifier': 'HOUR', 'symbol': 'h', 'origin': 'Non-SI unit mentioned in the SI', 'flg_available': True, 'description': 'Unit of time'}, {'id': 36, 'name': 'Day', 'identifier': 'DAY', 'symbol': 'd', 'origin': 'Non-SI unit mentioned in the SI', 'flg_available': True, 'description': 'Unit of time'}, {'id': 37, 'name': 'Year', 'identifier': 'YEAR', 'symbol': 'a', 'origin': 'Other', 'flg_available': True, 'description': 'Unit of time (equal to either 365 or 366 days)'}, {'id': 38, 'name': 'Bar', 'identifier': 'BAR', 'symbol': 'bar', 'origin': 'Common unit not officially sanctioned', 'flg_available': True, 'description': 'Metric unit of pressure'}, {'id': 39, 'name': 'Pixel', 'identifier': 'PIXEL', 'symbol': 'px', 'origin': 'Other', 'flg_available': True, 'description': 'Unit of programmable color on a computer display or in a computer image'}, {'id': 40, 'name': 'Byte', 'identifier': 'BYTE', 'symbol': 'B', 'origin': 'Other', 'flg_available': True, 'description': 'Unit of information'}, {'id': 41, 'name': 'Bit', 'identifier': 'BIT', 'symbol': 'bit', 'origin': 'Other', 'flg_available': True, 'description': 'Unit of information'}, {'id': 42, 'name': 'Metre per second', 'identifier': 'METER_PER_SECOND', 'symbol': 'm/s', 'origin': 'SI derived unit', 'flg_available': True, 'description': 'SI derived unit of both speed (scalar) and velocity (vector quantity which specifies both magnitude and a specific direction), defined by distance in metres divided by time in seconds.'}, {'id': 43, 'name': 'Volts per second', 'identifier': 'VOLT_PER_SECOND', 'symbol': 'V/s', 'origin': 'SI derived unit', 'flg_available': True, 'description': 'SI derived rate unit of change of Voltage'}, {'id': 44, 'name': 'Ampere per second', 'identifier': 'AMPERE_PER_SECOND', 'symbol': 'A/s', 'origin': 'SI derived unit', 'flg_available': True, 'description': 'Unit of electrical charge'}, {'id': 45, 'name': 'Percentage', 'identifier': 'PERCENT', 'symbol': '%', 'origin': 'Other', 'flg_available': True, 'description': ''}, {'id': 46, 'name': 'Unknown', 'identifier': 'NOT_ASSIGNED', 'symbol': 'N_A', 'origin': 'Other', 'flg_available': True, 'description': ''}]


create_parameter_api_response = {'id': 1234, 'data_source': 'LOCATION/CLASS/PARAM/00', 'name': 'Test device', 'value': 111.6604995727539, 'minimum': 106.80351257324219, 'maximum': 116.29515075683594, 'mean': 111.57698059082031, 'standard_deviation': 2.211357831954956, 'data_type_id': 20, 'parameter_type_id': 1, 'unit_id': 12, 'unit_prefix': '', 'flg_available': True, 'description': 'test device', 'data_groups_parameters': []}



import os
from pathlib import Path
from utils.utils import flatten_list


def MSG_FILE_NOT_FOUND(name, obj):
    return f'{name} does not exist or it is not a directory: {obj}'


class InputValidator:
    INVALID_VALUE = -1

    @staticmethod
    def validate_input(base_folder, scope, proposal, run=None, files=[], **kwargs):

        data_view = {'scope': scope}

        # If the base folder is not specified then use the current working directory,
        # Otherwise make sure that the base folder exist and is a folder
        base_folder = Path(base_folder or os.getcwd()).resolve()
        if not base_folder.is_dir():
            raise FileNotFoundError(MSG_FILE_NOT_FOUND('Base folder', base_folder))

        data_view["base_folder"] = str(base_folder)

        # Validata the proposal information
        proposal = InputValidator.validate_numerical_input(proposal, 'p')

        if proposal == InputValidator.INVALID_VALUE:
            raise ValueError(f'Invalid proposal number')

        # Construct data path
        if proposal is not None:
            # proposal = f'p{proposal}'
            data_path = base_folder / f'p{proposal}' / 'raw'

            # Make sure that the proposal folder exists
            if not data_path.is_dir():
                raise FileNotFoundError(MSG_FILE_NOT_FOUND('Proposal folder', data_path))

            data_view["proposal_number"] = proposal
        else:
            data_path = base_folder

        data_view["data_path"] = str(data_path)

        # Validate and construct runs list
        if run is not None:
            # Validate the run information
            run = InputValidator.validate_numerical_input(run, 'r')

            if run == InputValidator.INVALID_VALUE:
                raise ValueError(f'Invalid run number')

            path_list = [data_path / f'r{run:04}']
        else:
            # Scan all run folders within the data path
            path_list = [f for f in data_path.iterdir()
                         if InputValidator.validate_numerical_input(f.stem, 'r') != InputValidator.INVALID_VALUE]

        data_view['runs'] = [{'run_number': path.stem,
                              'path': str(path),
                              'files': InputValidator.discover_run_data(path)
                              } for path in path_list if path.is_dir()]

        # Make sure that all specified files exists
        # Filter and remove duplicate file paths
        data_view['files'] = list(set(f for f in files if Path(f).is_file())) if files is not None else []

        return data_view

    @staticmethod
    def discover_all_run_data(data_view):
        for run in data_view['runs']:
            run['data_files'] = [str(f) for f in list(Path(f'{run["path"]}').rglob("*.h5"))]

    @staticmethod
    def discover_run_data(path, pattern='*.h5'):
        return [str(f) for f in list(Path(path).rglob(pattern))]

    @staticmethod
    def validate_numerical_input(val, prefix='', type=int):
        """
        Validates and extracts an integer input after removing a specified prefix.

        Parameters:
        - val: The input value to be validated (can be any type).
        - prefix: The prefix to be removed from the input value (default is an empty string).

        Returns:
        - An integer if the modified input consists only of digits, otherwise returns None.
        """
        if val is None:
            return None
        # Convert the input value to a string
        tmp = str(val)

        # Check if the string starts with the specified prefix
        if tmp.startswith(prefix):
            # If it does, remove the prefix
            tmp = tmp[len(prefix):]

        # Check if the modified string consists only of digits
        return type(tmp) if tmp.isdigit() else InputValidator.INVALID_VALUE



from loguru import logger

# Check if required libraries are imported
try:
    import pandas as pd
except ImportError:
    raise ImportError("Pandas library is required for this function.")

from utils.defines import Scope
from utils.Trace import enable_trace


class MetadataExtractor:

    def __init__(self, operators) -> None:
        self.operators = operators

    @enable_trace
    def process(self, data_view):

        #  trace(f'Data view : {data_view}')

        scope = data_view.get('scope', Scope.PROPOSAL)

        match scope:
            case Scope.RUN:
                runs = data_view['runs']
                for run in runs:
                    data = run['data']
                    run['parameters'] = MetadataExtractor.describe_data(data, self.operators)
                    del run['data']
            case Scope.FILE:
                data_files = data_view['files']
                for file in data_files:
                    data = file['data']
                    file['parameters'] = MetadataExtractor.describe_data(data, self.operators)
                    del file['data']

            case _:  # Scope.PROPOSAL
                data = data_view['data']
                data_view['parameters'] = MetadataExtractor.describe_data(data, self.operators)
                del data_view['data']

        return data_view

    @staticmethod
    def describe_data(datasets, operators):
        """
        Describe the given datasets and return the result as a dictionary.

        Parameters:
        - datasets (dict): A dictionary containing datasets, where keys are paths and values
                        are dictionaries with 'data' (numpy array) and 'metadata' keys.

        Returns:
        - dict: A dictionary with paths as keys and values as dictionaries containing
                'data' (descriptive statistics) and 'metadata'.
        """

        result = {}

        for path, data in datasets.items():
            df_describe = pd.DataFrame(data['data'])
            df_describe.dropna(inplace=True)
            percentiles = [.50]

            # Generate descriptive statistics and update the result dictionary
            desc = df_describe.describe(percentiles=percentiles).to_dict()[0]
            desc['med'] = desc['50%']
            del desc['50%']

            summary = {k: v for k, v in desc.items() if k in operators}

            result[path] = {'data': summary, 'attributes': data['attributes'], 'run_number': data['run_number'], 'proposal_number': data['proposal_number']}

        return result



from loguru import logger
from load.ServiceInterface import ServiceInterface
from requests.auth import HTTPBasicAuth
import urllib3

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


def retrieve_proposal_info_by_number(service_interface, proposal_numer):
    response = service_interface.api_get(f'proposals/by_number/{proposal_numer}', {})  # {'data_source': "data_source", 'name': "name"}) #, {'proposal_id': 4})
    return response.json()


def retrieve_proposal_info_by_id(service_interface, proposal_id):
    response = service_interface.api_get(f'proposals/{proposal_id}', {})  # {'data_source': "data_source", 'name': "name"}) #, {'proposal_id': 4})
    return response.json()


def retrieve_run_info(service_interface, proposal_id, run_number):
    response = service_interface.api_get(f'proposals/by_number/{proposal_id}/runs/{run_number}', {})  # {'data_source': "data_source", 'name': "name"}) #, {'proposal_id': 4})
    return response.json()


def retrieve_run_info_by_id(service_interface, run_id):
    response = service_interface.api_get(f'runs/{run_id}', {})  # {'data_source': "data_source", 'name': "name"}) #, {'proposal_id': 4})
    return response.json()


def fetch_all_parameters(service_interface):
    response = []


# /dev_metadata/proposals/:proposal_id/runs/:id/parameters
def fetch_all_records(service_interface, api_id):
    response = []

    for page in range(1, 100):
        res = service_interface.api_get(api_id, {'page': page, 'page_size': -1})

        try:
            lst = res.json()
            if isinstance(lst, list) and len(lst):
                response.extend(lst)
            else:
                break
        except (ValueError, TypeError):
            break

    return response


def extract_parameters_with_data_groups(parameters, key="data_groups_parameters"):
    return [param for param in parameters if isinstance(param[key], list) and len(param[key])]


def extract_parameters_by_data_group_id(parameters, data_group_id, key="data_groups_parameters"):
    return [param for param in parameters if isinstance(param[key], list) and len(param[key]) > 0 and param[key][0]['data_group_id'] == data_group_id]


def extract_files_by_data_group_id(files, data_group_id, key="data_group_id"):
    return [file for file in files if file[key] == data_group_id]


def retrieve_units(service_interface):
    response = service_interface.api_get(f'units')
    return response.json()


def retrieve_unit_info(service_interface, unit_id):
    response = service_interface.api_get(f'units/{unit_id}')
    return response.json()


def inject_parameter(service_interface, parameter):
    response = service_interface.api_post('parameters', parameter)
    # print(response)
    return response.json()


def retrieve_parameter(service_interface, parameter_id):
    # response = service_interface.api_get(f'../proposals/{proposal_id}/runs/{run_id}/data_groups', {'run_id': run_id}) # {'data_source': "data_source", 'name': "name"}) #, {'proposal_id': 4})
    response = service_interface.api_get(f'parameters/{parameter_id}')  # {'data_source': "data_source", 'name': "name"}) #, {'proposal_id': 4})
    return response.json()


def update_parameter(service_interface, parameter_id, parameter):
    # response = service_interface.api_get(f'../proposals/{proposal_id}/runs/{run_id}/data_groups', {'run_id': run_id}) # {'data_source': "data_source", 'name': "name"}) #, {'proposal_id': 4})
    response = service_interface.api_put(f'parameters/{parameter_id}', parameter)  # {'data_source': "data_source", 'name': "name"}) #, {'proposal_id': 4})
    return response.json()


def delete_parameter(service_interface, parameter_id):
    response = service_interface.api_delete(f'parameters', {'id': parameter_id})  # {'data_source': "data_source", 'name': "name"}) #, {'proposal_id': 4})
    return response  # .json()


def inject_run_paramter(service_interface, run_id, parameter_id):
    return {}


def get_proposal_details(proposal_number):
    pass


def retrieve_data_group(service_interface, data_group_id):
    response = service_interface.api_get(f'data_groups/{data_group_id}')  # {'data_source': "data_source", 'name': "name"}) #, {'proposal_id': 4})
    return response.json()


def connect_to_service(app_info, user_info):
    service_interface = ServiceInterface(app_info['id'], app_info['secret'], user_info['email'], app_info['url'], app_info['scope'])

    auth = HTTPBasicAuth(app_info['id'], app_info['secret'])

    return service_interface
    req_proposal = {
        'number': 990390
    }

    response = service_interface.api_get(f'proposals/by_number/{2}', {})  # {'data_source': "data_source", 'name': "name"}) #, {'proposal_id': 4})
    logger.info(response.content)
    trace(json.dumps(json.loads(response.text), indent=4))

    response = service_interface.api_get(f'proposals/by_number/{2}/runs', {})  # {'data_source': "data_source", 'name': "name"}) #, {'proposal_id': 4})
    logger.info(response.content)
    trace(json.dumps(json.loads(response.text), indent=4))

    response = service_interface.api_get(f'proposals/by_number/{2}/runs/{3}', {})  # {'data_source': "data_source", 'name': "name"}) #, {'proposal_id': 4})
    logger.info(response.content)
    trace(json.dumps(json.loads(response.text), indent=4))

    parameter = {
        'parameter': {
            'data_source': 'MID/XTD9/XGM/MAIN/DESY',
            'name': 'PhotonFlux',
            'value': 15.6,
            'minimum': 35.0,
            'maximum': 106.0,
            'mean': 45.5,
            'standard_deviation': 7.5,
            'data_type_id': 20,
            'parameter_type_id': 1,
            'unit_id': 1,
            'unit_prefix': '',
            'flg_available': True,
            'description': 'desc 01'
        }
    }

    response = service_interface.api_post('parameters', parameter)  # {'data_source': "data_source", 'name': "name"}) #, {'proposal_id': 4})
    logger.info(response.content)
    trace(json.dumps(json.loads(response.text), indent=4))

    response = service_interface.api_get(f'parameters', {})  # {'data_source': "data_source", 'name': "name"}) #, {'proposal_id': 4})
    logger.info(response.content)
    trace(json.dumps(json.loads(response.text), indent=4))

    return

    data_group_type = {
        'data_group_type': {
            'name': "Gamma",
            'identifier': "GAMMA",
            'flg_available': 'false',
            'description': 'desc 01'
        }
    }

    response = service_interface.api_post('data_group_types', data_group_type)  # {'data_source': "data_source", 'name': "name"}) #, {'proposal_id': 4})
    logger.info(response.content)

    json_obj = json.loads(response.text) if len(response.text) else {}
    logger.info(json_obj)
    data_group_type_id = json_obj['id']
    # data_group_type_id = 156
    logger.info(f"Delete existing data group type: {data_group_type_id}.")

    response = service_interface.api_delete('data_group_types', {"id": data_group_type_id})  # {'data_source': "data_source", 'name': "name"}) #, {'proposal_id': 4})

    json_obj = json.loads(response.text) if len(response.text) else {}
    logger.info(json_obj)
    # logger.info (response.content)

    curr_method_name = inspect.currentframe().f_code.co_name
    trace(f'{MODULE_NAME}.{curr_method_name} ==> {response}')

    resp = handle_response(response, GET, OK, MODULE_NAME)

    json_obj = json.loads(response.text) if len(response.text) else {}

    logger.info(json_obj)

    logger.info(('resp: {0}'.format(resp)))



from http import HTTPStatus
from oauthlib.oauth2 import BackendApplicationClient
from requests_oauthlib.oauth2_session import OAuth2Session
from utils.defines import MAX_RETRIES, TIMEOUT, SSL_VERIFY
from utils.Trace import enable_trace
import json
import inspect
import requests


class ServiceInterface:

    def __init__(self, client_id, client_secret, user_email, base_url, scope) -> None:

        self.client_id = client_id
        self.client_secret = client_secret
        self.scope = scope
        self.user_email = user_email
        vars(self).update(ServiceInterface.build_app_urls(base_url))

        self.fetch_token_kwargs = ServiceInterface.build_fetch_token_kwargs(self.token_url, self.client_secret)
        self.headers = ServiceInterface.build_default_headers(self.user_email)

        self.oauth_client = BackendApplicationClient(self.client_id)

        self.oauth_session = OAuth2Session(client=self.oauth_client, scope=self.scope, auto_refresh_kwargs=self.fetch_token_kwargs)

        # res = ServiceInterface.get_access_token("https://192.168.56.101:3000/dev_metadata/oauth/token", self.client_id, self.client_secret)
        # assert (res.get("expires_in", None))

    @staticmethod
    def get_access_token(url, client_id, client_secret):
        response = requests.post(
            url,
            data={"grant_type": "client_credentials"},
            auth=(client_id, client_secret),
            verify=False
        )
        return response.json()

    @staticmethod
    def build_fetch_token_kwargs(token_url, client_secret):
        return {
            'token_url': token_url,
            'client_secret': client_secret,
            'max_retries': MAX_RETRIES,
            'timeout': TIMEOUT,
            'verify': SSL_VERIFY
        }

    @staticmethod
    def build_app_urls(base_app_url):
        __suffix = {
            'base': '/',
            'token': 'oauth/token',
            'refresh': 'oauth/token',
            'authorize': 'oauth/authorize',
            'api': 'api',
        }
        base_url = base_app_url.rstrip('/')

        # dict([(x,y) for y,x in enumerate('cbad')])

        # return dict([{key : base_url + '/' +value} for key, value in __suffix])
        result_dict = {key + '_url': (base_url + '/' + value) for key, value in __suffix.items()}
        return result_dict

    @staticmethod
    def build_default_headers(client_email):
        return {
            'content-type': 'application/json',
            'Accept': 'application/json; version=1',
            'X-User-Email': client_email,
        }

    @enable_trace
    def get_session(self):
        self.oauth_session.fetch_token(**self.fetch_token_kwargs)
        return self.oauth_session

    @enable_trace
    def api_get(self, api_id, parameters={}):
        api_url = self.api_url + '/' + api_id + '/'
        kwargs = {'params': {'page': 1, 'page_size': -1}.update(parameters), 'allow_redirects': True}
        kwargs = {'params': parameters, 'allow_redirects': True}
        return self.get_session().get(api_url, headers=self.headers, **kwargs, verify=SSL_VERIFY)

    @enable_trace
    def api_post(self, api_id, parameters={}):
        api_url = self.api_url + '/' + api_id + '/'

        kwargs = {'allow_redirects': True, "data": json.dumps(parameters), 'params': {'page': 1, 'page_size': -1, }}
        # kwargs = json.dumps(kwargs)
        # 'data_group_type': parameters |
        trace(f'API URL: {api_url}')
        trace(f'Input arguments: {parameters}')
        trace(f'POST query arguments: {kwargs}')

        return self.get_session().post(api_url, headers=self.headers, **kwargs, verify=SSL_VERIFY)

    @enable_trace
    def api_put(self, api_id, parameters={}):
        api_url = self.api_url + '/' + api_id + '/'
        kwargs = {'params': {'page': 1, 'page_size': -1}, "data": json.dumps(parameters), 'allow_redirects': True}
        return self.get_session().put(api_url, headers=self.headers, **kwargs, verify=SSL_VERIFY)

    @enable_trace
    def api_delete(self, api_id, parameters={}):
        api_url = self.api_url + '/' + api_id + '/' + str(parameters["id"])
        kwargs = {'allow_redirects': True, "data": json.dumps(parameters), 'params': {'page': 1, 'page_size': -1, }}
        return self.get_session().delete(api_url, headers=self.headers, **kwargs, verify=SSL_VERIFY)


def handle_response(response, action, success_code, module_name):
    content = response_to_json(response)
    pagination = load_json_from_headers(response)

    if response.status_code == HTTPStatus.NO_CONTENT and not content:
        formatted_response = response_success(module_name, action, pagination, content)
    elif response.status_code == success_code:
        if content or content == []:
            formatted_response = response_success(module_name, action, pagination, content)
        else:
            formatted_response = response_error(module_name, action, content)
    else:
        app_info = content.get('info', f'HTTP request status code: {response.status_code}')
        formatted_response = response_error(module_name, action, app_info)

    return formatted_response


def string_to_json(hash_str):
    return {} if hash_str in ['', '[]'] else json.loads(hash_str)


def response_to_json(response):
    return string_to_json(response.content.decode('utf8'))


def load_json_from_headers(response):
    headers = {}
    if response.headers:
        headers['Date'] = response.headers.get('Date')
        headers['X-Total-Pages'] = response.headers.get('X-Total-Pages')
        headers['X-Count-Per-Page'] = response.headers.get('X-Count-Per-Page')
        headers['X-Current-Page'] = response.headers.get('X-Current-Page')
        headers['X-Total-Count'] = response.headers.get('X-Total-Count')
    return headers


class ActionType:
    CREATE = 'CREATE'
    UPDATE = 'UPDATE'
    GET = 'GET'
    DELETE = 'DELETE'
    SET = 'SET'


def response_success(module_name, action, r_pagination, r_content):
    action_messages = {
        ActionType.CREATE: f'{module_name} created successfully',
        ActionType.UPDATE: f'{module_name} updated successfully',
        ActionType.GET: f'Got {module_name} successfully',
        ActionType.DELETE: f'{module_name} deleted successfully',
        ActionType.SET: f'{module_name} set successfully',
    }

    msg = action_messages.get(action, '')
    if not msg:
        return response_error(module_name, action, r_content)

    res = {
        'success': True,
        'info': msg,
        'app_info': {},
        'pagination': r_pagination,
        'data': r_content
    }

    return res


def response_error(module_name, action, app_info):
    action_messages = {
        ActionType.CREATE: f'Error creating {module_name}',
        ActionType.UPDATE: f'Error updating {module_name}',
        ActionType.GET: f'Error getting {module_name}',
        ActionType.DELETE: f'Error deleting {module_name}',
        ActionType.SET: f'Error setting {module_name}',
    }

    msg = action_messages.get(action, 'ACTION is not correct!')
    res = {
        'success': False,
        'info': msg,
        'app_info': app_info,
        'pagination': {},
        'data': {}
    }

    return res




import unittest

from utils.utils import create_parameter_data, create_metadata_client, fetch_proposal_details, fetch_run_details
from unittest.mock import mock_open, patch
from .secrets import *
from metadata_api import MetadataClient
from loguru import logger
from .mdc_api_responses_mock import *
import json

import unittest
from unittest.mock import MagicMock, patch

# Import the function from your module
# from your_module import fetch_proposal_details


class TestFetchProposalDetails(unittest.TestCase):

    def setUp(self):
        """Set up common test data."""
        self.mock_client = MagicMock()  # Mock the client connection
        self.client_api = MagicMock()  # create_metadata_client(APP_INFO, USER_INFO)
        self.proposal_number = 900390
        self.run_number = 704

    def test_fetch_proposal_details_valid(self):
        """Test fetch_proposal_details with a valid proposal response."""
        proposal_number = "12345"
        expected_response = proposal_info_response  # {"id": 1, "name": "Test Proposal", "status": "Approved"}

        # Mock the API call response
        self.mock_client.get_proposal_by_number_api.return_value.json.return_value = expected_response

        result = fetch_proposal_details(self.mock_client, proposal_number)

        self.mock_client.get_proposal_by_number_api.assert_called_once_with(proposal_number)
        self.assertEqual(result, expected_response)

    def test_fetch_proposal_details_valid_api(self):
        """Test fetch_proposal_details with a valid proposal response."""
        # Mock the API call response
        res_api = self.client_api.get_proposal_by_number_api(self.proposal_number).json()

        self.proposal_details = fetch_proposal_details(self.client_api, self.proposal_number)
        self.assertIsNotNone(self.proposal_details)

        self.assertEqual(res_api, self.proposal_details)

    def test_fetch_proposal_details_valid_api(self):
        expected_response = run_info_response
        self.mock_client.get_runs_by_proposal_number_api.return_value.json.return_value = proposal_runs_response
        self.mock_client.get_run_by_id_api.return_value.json.return_value = run_info_response
        self.run_info = fetch_run_details(self.mock_client, self.proposal_number, self.run_number)
        self.assertIsNotNone(self.run_info)
        self.assertEqual(self.run_info, expected_response)

        return
        self.mock_client.get_proposal_by_number_api.assert_called_once_with(proposal_number)
        self.assertEqual(result, expected_response)

    # @patch('loguru.logger.error')  # Mock logger to prevent real logging
    # def test_fetch_proposal_details_missing_id(self, mock_error):
    #     """Test when 'id' is missing from the API response."""
    #     proposal_number = "67890"
    #     error_response = {"info": f"Proposal number {proposal_number} not found"}

    #     # Mock API call returning an invalid response (missing 'id')
    #     self.mock_client.get_proposal_by_number_api.return_value.json.return_value = error_response
    #     # ex = ValueError(error_response["info"])

    #     # with self.assertRaises(ValueError) as context:
    #     result = fetch_proposal_details(self.mock_client, proposal_number)

    #     #self.assertEqual(str(context.exception), error_response["info"])
    #     self.mock_client.get_proposal_by_number_api.assert_called_once_with(proposal_number)
    #     mock_error.assert_called_once()
    #     self.assertIsNone(result)

    @patch('loguru.logger.error')  # Mock logger to prevent real logging
    def test_fetch_proposal_details_exception_handling(self, mock_error):
        """Test fetch_proposal_details when an exception occurs."""
        proposal_number = "9999999"

        error_response = {"info": f"Proposal number {proposal_number} not found"}

        # Mock API call raising an exception, proposal not found
        self.mock_client.get_proposal_by_number_api.side_effect = ValueError(error_response["info"])

        result = fetch_proposal_details(self.mock_client, proposal_number)

        self.mock_client.get_proposal_by_number_api.assert_called_once_with(proposal_number)
        # mock_error.error.assert_called_once()
        print("Calls are: ", mock_error.mock_calls)

        calls = mock_error.mock_calls
        assert len(calls) == 1
        # Unpack call
        function_called, args, kwargs = calls[0]
        # assert(function_called == "error")
        value_error = args[0]

        assert isinstance(value_error, ValueError)
        assert value_error.args[0] == error_response["info"]

        self.assertIsNone(result)


class TestParameterCRUD(unittest.TestCase):

    def setUp(self):
        self.mock_client_api = MagicMock()
        self.client_api = create_metadata_client(APP_INFO, USER_INFO)

    def tearDown(self):
        pass  # cls.client_api.destroy()

    @staticmethod
    def create_dummy_parameter(key):
        data = {'count': 6000.0,
                'mean': 111.57698059082031,
                'std': 2.211357831954956,
                'min': 106.80351257324219,
                'max': 116.29515075683594,
                'med': 111.6604995727539}

        attrs = {
            'device': key,
            'name': 'Test device',
            'unit_id': '12',
            'unit_prefix_name': '',
            'alias': "test device",
        }

        return create_parameter_data(key, data, attrs)

    @patch("loguru.logger.warning")
    @patch("loguru.logger.error")
    def test_inject_parameter(self, mock_error, mock_warning):
        self.client_api = create_metadata_client(APP_INFO, USER_INFO)
        self.assertIsNotNone(self.client_api)  # The result should be None because of the error
        param = self.create_dummy_parameter("LOCATION/CLASS/PARAM/00")
        expected_result = {"id": int()} | param

        self.mock_client_api.get_all_units_api.return_value.json.return_value = get_all_units_reponse
        units = self.client_api.get_all_units_api(0).json()
        self.assertIsNotNone(units)  # The units should not be None

        self.assertIsNotNone(param)  # The result should be None because of the error

        self.mock_client_api.create_parameter_api.return_value.json.return_value = expected_result
        res = self.client_api.create_parameter_api(param).json()
        print(json.dumps(res, indent=4))
        print(json.dumps(param, indent=4))

        self.assertIsNotNone(res)  # The result should be None because of the error
        mock_warning.assert_not_called()  # No warning should be logged
        mock_error.assert_not_called()  # An error should be logged

        assert 'id' in res
        param_id = res['id']
        res = self.client_api.get_parameter_by_id_api(param_id)
        res = self.client_api.delete_parameter_api(param_id)
        res = self.client_api.get_parameter_by_id_api(param_id)


if __name__ == "__main__":
    unittest.main()



from copy import deepcopy
from loguru import logger
import h5py as h5
import numpy as np

from utils.defines import Scope
from utils.Trace import enable_trace


class DataLoader:

    def __init__(self, data_view):
        self.data_view = data_view
        self.scope = data_view.get('scope', Scope.PROPOSAL)

    @enable_trace
    def process(self, data_parameters, data_attributes):
        """
        Load specified parameters and attributes for the data_view provided above.

        First, we validates the parameters, load the data and processes it accordingly
        based on the specified scope (RUN, FILE, or PROPOSAL).

        Parameters:
        - data_parameters (dict): Parameters specifying the list of parameters to extract from the files.
        - data_attributes (dict): Additional attributes associated with the data to be extracted also.

        Returns:
        dict: The raw data associated with each member of the list of data_parameters.

        """

        #  trace(f'Data parameters specs: {data_parameters}')

        validated_data_parameters = DataLoader.validate_data_parameters(data_parameters)
        logger.debug(f'Load data parameters: {validated_data_parameters}')

        match self.scope:
            case Scope.RUN:
                runs = self.data_view['runs']
                for run in runs:
                    data_files = run['files']
                    run['data'] = DataLoader.batch_process(data_files, validated_data_parameters, data_attributes)

            case Scope.FILE:
                data_files = self.data_view['files']
                for file in data_files:
                    file['data'] = DataLoader.process_file(file['file'], validated_data_parameters, data_attributes)

            case _:  # Scope.PROPOSAL
                data_files = self.data_view['files']
                self.data_view['data'] = DataLoader.batch_process(data_files, validated_data_parameters, data_attributes)

        return self.data_view

    @staticmethod
    def validate_data_parameters(data_parameters):
        required_keys = ['section', 'group', 'value']
        result = []

        for param in data_parameters:
            if not all(name in param for name in required_keys):
                raise ValueError(f"Wrong description of the parameter: {param}")

            value = param['value']

            if isinstance(value, str):
                result.append(param)
            elif isinstance(value, list):
                result.extend(deepcopy(param) | {'value': v} for v in value)
            else:
                raise TypeError(f"Value type error: {value}")

        return result

    @staticmethod
    def create_data_path(param_desc, has_value_suffix=True):
        """
        Create a data path based on the given parameter description.

        Parameters:
        - param_desc (dict): Dictionary containing 'section', 'group', and 'value'.
        - has_value (bool): Indicates whether to include '/value' as prefix to the path.

        Returns:
        - str: Constructed data path.
        """
        required_keys = ['section', 'group', 'value']
        if not all(name in param_desc for name in required_keys):
            raise ValueError("Wrong description of the parameter")

        section = param_desc['section']
        group = param_desc['group']
        value = param_desc['value']

        # value_suffix = '/value' if has_value_suffix else ''
        # return '{}/{}/{}/{}{}'.format(section, path, group, property_name, value_suffix)
        return f'{section}/{group}/{value}' + ('/value' if has_value_suffix else '')

    @staticmethod
    def process_file(filename, data_params, data_attributes):
        """
        Extract dataset from an HDF5 file based on the provided data parameters.

        Parameters:
        - filename (str): Path to the HDF5 file.
        - data_params (list): List of parameter descriptions for creating data paths.
        - data_attributes (list): List of attributes keys to extract for each dataset.

        Returns:
        - dict: Dictionary containing data paths, datasets, and attributes.
        """
        result = {}

        with h5.File(filename, 'r') as f:
            run_number = f["METADATA/runNumber"][()][0]
            proposal_number = f["METADATA/proposalNumber"][()][0]
            for param in data_params:
                try:
                    data_path = DataLoader.create_data_path(param)
                    if data_path not in f:
                        logger.warning(f"Dataset not found: {data_path}")
                        continue

                    dataset = f[data_path][()]

                    # Check if attributes exist before accessing
                    all_attrs = f[data_path].attrs if hasattr(f[data_path], 'attrs') else {}

                    # Extract dataset attribues based on data_attributes
                    attributes = {k: v for k, v in all_attrs.items() if k in data_attributes}

                    for k, v in {"section": "section", "device": "group", "name": "value"}.items():
                        attributes[k] = param[v]

                    #  trace(attributes)
                    # Update the result dictionary with both data and attributes
                    result[data_path] = {'data': dataset, 'attributes': attributes, 'run_number': run_number, 'proposal_number': proposal_number}
                except Exception as error:
                    logger.error(f'{error}. Dataset: {data_path}')

        return result

    @staticmethod
    def batch_process(data_files, data_params, data_attributes):
        """
        Aggregate datasets from multiple HDF5 files.

        Parameters:
        - data_files (list): List of file paths.
        - data_params (dict): List of parameters.
        - data_attributes (list): List of attributes to extract.

        Returns:
        - dict: Aggregated datasets.
        """
        # Use a list for temporary storage
        data_list = {}

        for param in data_params:
            data_path = DataLoader.create_data_path(param)
            data_list[data_path] = {'data': [], 'attributes': {}, 'run_number': None, 'proposal_number': None}

        for filename in data_files:
            data = DataLoader.process_file(filename, data_params, data_attributes)

            for p, v in data.items():
                data_list[p]['data'].append(v['data'])
                data_list[p]['attributes'] = v['attributes']
                data_list[p]['run_number'] = v['run_number']
                data_list[p]['proposal_number'] = v['proposal_number']

        result = {}
        for path, data in data_list.items():
            if not data['data']:
                continue
            aggregated_data = np.concatenate(data['data'])
            result.update({path: {'data': aggregated_data, 'attributes': data['attributes'], 'run_number': int(data['run_number']), 'proposal_number': int(data['proposal_number'])}})

        return result



from utils.defines import Scope


class InputProcessor:

    @staticmethod
    def process_input(data_view):

        scope = data_view["scope"]

        data_files = data_view['files']

        if scope != Scope.RUN:
            if 'runs' not in data_view:
                raise KeyError('Missing mandaroty field "runs"')
            # Merge all the run files under the proposal
            for run in data_view['runs']:
                data_files.extend(run['files'])
                run['files'] = []

            data_view['files'] = list(set(data_files))
            data_view['files'].sort()

            del data_view['runs']
        else:
            if 'files' in data_view:
                del data_view['files']

        match scope:
            case Scope.FILE:
                if data_files is not None:
                    data_files = [{'file': f, "parameters": {}} for f in data_files]
                    data_view.update({'files': data_files})
            case Scope.RUN:
                runs = data_view['runs'] if 'runs' in data_view else None
                if type(runs) is not list:
                    raise TypeError('List of runs is not well-formet')
                for run in runs:
                    run.update({'parameters': {}})
            case _:  # Scope.PROPOSAL
                data_view.update({'parameters': {}})

        return data_view



#!/usr/bin/python3

from utils.CommandLine import process_arguments
from extract.InputValidator import InputValidator
from extract.InputProcessor import InputProcessor
from extract.DataLoader import DataLoader
from transform.MetadataExtractor import MetadataExtractor
from utils.Trace import Trace, enable_trace
from utils.utils import load_data_from_file, resolve_unit_info, create_parameter_data, create_metadata_client, load_configuration
from utils.defines import STAT_OPERATORS
from loguru import logger
import json
from utils.DataView import H5FileView, H5JsonEncoder

from metadata_api import MetadataClient
from utils.defines import MAX_RETRIES, TIMEOUT, SSL_VERIFY

PROC_STAGE = ['config', 'extract', 'transform', 'load']


@enable_trace
def main(args=None):

    # Process the list of the command line arguments
    arguments = process_arguments()

    # Setup logger and trace
    logger.debug(f'Command line arguments: {arguments}')
    trace = Trace().setup(False)  # arguments.verbose

    proc_stage = arguments.stage
    try:
        app_info, user_info = load_configuration()
        if not (user_info and app_info):
            raise ValueError('Invalid application or user information')

        logger.info(f'Application info: {json.dumps(app_info, indent=4)}')
        logger.info(f'User info: {json.dumps(user_info, indent=4)}')

        logger.info(f'Connecting to Metadata service {app_info["url"]} ...')
        client_conn = create_metadata_client(app_info, user_info)
    except Exception as e:
        logger.error(e)
        return

    logger.info(f'Successfuly connected to Metadata service')

    # Dump data from MDC to JSON files
    #  dump_data_from_mdc(service_interface)

    proposal_number = arguments.proposal
    run_number = arguments.run

    try:
        proposal_info = client_conn.get_proposal_by_number_api(proposal_number).json()

        if 'id' not in proposal_info:
            raise ValueError(f"Unable to retrieve proposl info: {proposal_info}")

        proposal_id = proposal_info['id']
        logger.info(f'Proposal details by id ({proposal_id}):\n{json.dumps(proposal_info, indent=4)}')

        run_info = client_conn.get_runs_by_proposal_number_api(proposal_number, run_number).json()

        if 'runs' not in run_info:
            raise ValueError(run_info)  # ValueError(f"Unable to retrieve proposal info: {proposal_info}")

        run_id = run_info['runs'][0]['id']
        logger.info(f'Run details by id ({run_id}):\n{json.dumps(run_info, indent=4)}')

        try:
            logger.info("Retrieve run data using mdc client")
            run_info = client_conn.get_run_by_id_api(run_id).json()
            logger.info(f'Run details by id ({run_id}):\n{json.dumps(run_info, indent=4)}')
        except Exception as e:
            logger.error(e)

        logger.info(f'Run details by id:\n{json.dumps(run_info, indent=4)}')
        data_groups_repositories = run_info['data_groups_repositories']

        if isinstance(data_groups_repositories, list) and len(data_groups_repositories):
            data_group_id = data_groups_repositories[0]['id']

            logger.debug(f'Retrieve data group info {data_group_id}')
            data_group_info = client_conn.get_data_group_by_id_api(data_group_id).json()
            logger.info(f'Data group:\n{json.dumps(data_group_info, indent=4)}')

        data_group_parameters = []
        for id in data_group_info["parameter_ids"]:
            data_group_parameters.append(client_conn.get_parameter_by_id_api(id).json())

        data_files = client_conn.get_all_data_files_by_data_group_id_api(data_group_id).json()
        logger.debug(f"Data files:\n{json.dumps(data_files, indent=4)}")

        list_of_files = []
        for f in data_files:
            list_of_files.extend(eval(f['files']) if isinstance(f['files'], str) else f['files'])

        logger.info(f'Data files per group ({data_group_id}):\n{json.dumps(list_of_files, indent=4)}')

        prefix_path = data_group_info["prefix_path"]
        run_data_files = [prefix_path + f["relative_path"] for f in list_of_files]

        logger.info(f'Data files per run/data group ({run_number}/{data_group_id}):\n{json.dumps(run_data_files, indent=4)}')

    except Exception as e:
        logger.error(e)

    with open(arguments.data, 'r') as f:
        specs = json.load(f)

    if proc_stage == 'config':
        return

    logger.debug(f'Metadata specifications:\n{json.dumps(specs, indent=4)}')

    data_parameters = specs.get('parameters', {})
    data_attributes = specs.get('attributes', {})

    operators = set(specs.get('operators', STAT_OPERATORS))
    if "*" in operators:
        operators = STAT_OPERATORS

    parameters = InputValidator.validate_input(**dict(vars(arguments)))
    data_view = InputProcessor.process_input(parameters)

    logger.info(f'Data view: {json.dumps(data_view, indent=4)}')

    # ETL process for extracting metadata from HDF5 files

    # Extract (specific) data from files:
    data_loader = DataLoader(data_view)
    data = data_loader.process(data_parameters, data_attributes)

    # Transform data intp summary (metadata)
    metadata = MetadataExtractor(operators).process(data)

    logger.info(f'Metadata: {json.dumps(metadata, indent=4)}')

    if proc_stage == 'extract':
        return

    #  Load metadata into metadata catalog
    parameters = metadata['parameters']

    list_of_parameters = {}

    for key, value in parameters.items():
        attrs = value.get("attributes")
        list_of_parameters.update({attrs["device"] + '/' + attrs["name"]: key})

    logger.info(f"List of parameters from MDC: {json.dumps(data_group_parameters, indent=4)}")
    logger.debug(f"List of requested parameters: {json.dumps(list_of_parameters, indent=4)}")

    if proc_stage == 'transform':
        return

    units = client_conn.get_all_units_api(0).json()

    logger.info(f"Units:\n{units}")

    mdc_ops_summary = []

    for param in data_group_parameters:
        try:
            logger.debug(f'Process parameter:\n{json.dumps(param, indent=4)}')

            full_name = param["data_source"] + '/' + param["name"]
            if full_name in list_of_parameters:
                value = list_of_parameters[full_name]
                parameter_id = param["id"]
                param = parameters[value]
                data = param['data']
                attrs = param['attributes']

                unit_info = resolve_unit_info(units, attrs)
                attrs |= unit_info
                param = create_parameter_data(value, data, attrs)

                res = client_conn.update_parameter_api(parameter_id, param).json()
                if 'id' in res:
                    # logger.info(f'Updated parameter (id:{parameter_id}) successful:\n{json.dumps(res, indent=4)}')
                    mdc_ops_summary.append({"action": "update", "id": res["id"], "device": res["data_source"], "property": res["name"]})
                    list_of_parameters.pop(full_name)
                else:
                    raise RuntimeError(f'Update parameter (id:{parameter_id}) failed:\n{res}')
            else:
                trace(f"Delete parameter: {full_name}")
                parameter_id = param["id"]
                res = client_conn.delete_parameter_api(parameter_id)
                logger.info(f'Delete parameter ({param}):\n{res.status_code}')
                if res.status_code in [200, 204]:
                    mdc_ops_summary.append({"action": "delete", "id": param["id"], "device": param["data_source"], "property": param["name"]})
                else:
                    raise RuntimeError(f'Delete parameter (id:{parameter_id}) failed:\n{res}')
        except Exception as e:
            logger.warning(e)

    for key, value in list_of_parameters.items():
        param = parameters[value]
        trace(f"Add parameter: {key}\n{param}")

        data = param['data']
        attrs = param['attributes']
        unit_info = resolve_unit_info(units, attrs)
        attrs |= unit_info
        param = create_parameter_data(value, data, attrs)

        param['data_groups_parameters_attributes'] = [{'data_group_id': data_group_id}]
        # param['runs_parameters_attributes'] = [{'run_id': 2004}]

        res = client_conn.create_parameter_api(param).json()
        # logger.info(f'Insert parameter successful:\n{res}')
        mdc_ops_summary.append({"action": "create", "id": res["id"], "device": res["data_source"], "property": res["name"]})

    logger.info(f'Summary list of new/updated parameters:\n{json.dumps(mdc_ops_summary, indent=4)}')


if __name__ == "__main__":
    main()



import json
from loguru import logger
from mdc_api import fetch_all_records
from metadata_api import MetadataClient
from .defines import MAX_RETRIES, TIMEOUT, SSL_VERIFY
from .Trace import Trace, enable_trace

__all__ = ['flatten_list', 'setup_logger']


def flatten_list(nested_list):
    """
    Recursively flatten a nested list.

    Parameters:
    - nested_list: The input nested list to be flattened.

    Returns:
    - A flattened list containing all elements of the input nested list.
    """
    flattened_list = []
    for item in nested_list:
        if isinstance(item, list):
            # If the element is a list, recursively flatten it
            flattened_list.extend(flatten_list(item))
        else:
            # If the element is not a list, append it to the flattened list
            flattened_list.append(item)
    return flattened_list


def setup_logger(log_file_path='./etl.log', level="INFO",
                 format="{time:YYYY-MM-DD HH:mm:ss} | {level} | {message}"):
    if not hasattr(setup_logger, "logger"):
        logger.add(log_file_path, rotation="500MB", level=level, format=format)
        setup_logger.logger = True


# @enable_trace
def connect_to_database(databasename):
    import sqlite3
    con = sqlite3.connect(database=databasename)
    return con


# @enable_trace
def load_data_from_file(file_name):
    try:
        with open(file_name, 'r') as f:
            data = json.load(f)
            return data
    except Exception as error:
        logger.error(error)
        return None


def store_data_to_file(data, filename):
    with open(filename, 'w', encoding='utf-8') as f:
        json.dump(data, f, ensure_ascii=False, indent=4)


def dump_data_from_mdc(service_interface):

    tables_full_list = [
        'facilities',
        'units',
        'topics',
        'data_group_types',
        'data_types',
        'experiment_types',
        'parameter_types',
        'repositories',
        'instrument_types',
        'sample_types',
        'samples',
        'instrument_cycles',
        'experiments',
        'beamtimes',
        'proposals_users',
        'instruments',
        'data_groups',
        'data_files',
        'parameters',
        'runs',
        'proposals',
        'techniques',
        'proposals_techniques',
        'users'
    ]
    tables_names = [
    ]

    for table_name in tables_names:
        logger.info(f'Fetch all data from {table_name}.')
        try:
            data = fetch_all_records(service_interface, table_name)
            store_data_to_file(data, f'./resources/{table_name}.json')
        except Exception as e:
            logger.error(e)
    return


def resolve_unit_info(units, attrs):
    try:
        unit_name = attrs.get('unitName', 'Unknown')
        unit_symbol = attrs.get('unitSymbol', 'N_A')
        unit_prefix_name = attrs.get('metricPrefixName', '')
        unit_prefix_symbol = attrs.get('metricPrefixSymbol', '')
        unit_id = next(unit["id"] for unit in units if unit["symbol"] == unit_symbol)  # [key for key, value in my_dictionary.items() if value == 93]
    except Exception as e:
        unit_id = -1
        logger.warning(e)

    return {"unit_name": unit_name, "unit_symbol": unit_symbol, "unit_id": unit_id, "unit_prefix_name": unit_prefix_symbol, "unit_prefix_symbol": unit_prefix_symbol}


def create_metadata_client(app_info, user_info):
    BASE_URL = app_info.get("url")
    # Generate the connection (example with minimum parameter options)
    return MetadataClient(client_id=app_info.get("id"),
                          client_secret=app_info.get("secret"),
                          user_email=user_info.get("email"),
                          token_url=BASE_URL + '/oauth/token',
                          refresh_url=BASE_URL + '/oauth/token',
                          auth_url=BASE_URL + '/oauth/authorize',
                          scope=app_info.get("scope"),
                          base_api_url=BASE_URL + '/api/',
                          max_retries=MAX_RETRIES,
                          timeout=TIMEOUT,
                          session_token=None,
                          ssl_verify=False)


@enable_trace
def create_parameter_data(key, data, attrs):
    #  Example: {'count': 6000.0, 'mean': 111.57698059082031, 'std': 2.211357831954956, 'min': 106.80351257324219, 'max': 116.29515075683594, 'med': 111.6604995727539}
    try:
        param = {
            'data_source': attrs['device'],
            'name': attrs['name'],
            'value': data['med'],
            'minimum': data['min'],
            'maximum': data['max'],
            'mean': data['mean'],
            'standard_deviation': data['std'],
            'data_type_id': 20,
            'parameter_type_id': 1,
            'unit_id': attrs['unit_id'],
            'unit_prefix': attrs['unit_prefix_name'],
            'flg_available': True,
            'description': attrs.get('alias', ''),
        }

        if not (float(data['min']) <= float(data['mean']) <= float(data['max'])):
            logger.warning(f"Invalid parameter: {key}\n{data}")
            param['mean'] = data['med']

        return param
    except Exception as e:
        logger.error(e)
        return None


@enable_trace
def load_configuration():
    app_info = load_data_from_file('./resources/local_app.json')
    user_info = load_data_from_file('./resources/user.json')

    return app_info, user_info


@logger.catch
def load_config_and_connect_to_metadata_catalog():
    app_info, user_info = load_configuration()
    if not (user_info and app_info):
        raise ValueError('Invalid application or user information')

    logger.debug(f'Application info: {app_info}')
    logger.debug(f'User info: {user_info}')

    logger.info(f'Connecting to Metadata service {app_info["url"]} ...')
    client_conn = create_metadata_client(app_info, user_info)

    return client_conn


def fetch_proposal_details(client_conn, proposal_number):
    try:
        proposal_info = client_conn.get_proposal_by_number_api(proposal_number).json()

        if 'id' not in proposal_info:
            raise ValueError(proposal_info["info"])  # ValueError(f"Unable to retrieve proposal info: {proposal_info}")

        return proposal_info
    except Exception as e:
        logger.error(e)
        return None


def fetch_run_details(client_conn, proposal_number, run_number):
    try:
        run_info = client_conn.get_runs_by_proposal_number_api(proposal_number, run_number).json()

        if 'runs' not in run_info:
            raise ValueError(run_info)  # ValueError(f"Unable to retrieve proposal info: {proposal_info}")

        run_id = run_info['runs'][0]['id']
        logger.debug(f'Run details by number ({run_number}):\n{json.dumps(run_info, indent=4)}')

        run_info = client_conn.get_run_by_id_api(run_id).json()
        return run_info
    except Exception as e:
        logger.error(e)
        return None


def fetch_data_group_details(client_conn, data_group_id):  # data_groups_repositories
    try:
        # if isinstance(data_groups_repositories, list) and len(data_groups_repositories):
        #     data_group_id = data_groups_repositories[0]['id']

        logger.debug(f'Retrieve data group info {data_group_id}')
        return client_conn.get_data_group_by_id_api(data_group_id).json()
    except Exception as e:
        logger.error(e)
        return None


def fetch_parameters_details(client_conn, parameters_ids):
    try:
        return [client_conn.get_parameter_by_id_api(id).json() for id in parameters_ids]
    except Exception as e:
        logger.error(e)
        None


def fetch_all_data_group_files(client_conn, data_group_id, prefix_path):
    try:
        data_files = client_conn.get_all_data_files_by_data_group_id_api(data_group_id).json()
        logger.debug(f"Data files:\n{json.dumps(data_files, indent=4)}")

        list_of_files = []
        for f in data_files:
            list_of_files.extend(eval(f['files']) if isinstance(f['files'], str) else f['files'])

        logger.debug(f'Data files per group ({data_group_id}):\n{json.dumps(list_of_files, indent=4)}')

        return [prefix_path + f["relative_path"] + f["filename"] for f in list_of_files]
    except Exception as e:
        logger.error(e)
        return None


def execute_load_data_to_metadata_catalog(client_conn, data_group_id, list_of_parameters, data_group_parameters, parameters):
    units = client_conn.get_all_units_api(0).json()
    mdc_ops_summary = []
    for param in data_group_parameters:
        try:
            logger.debug(f'Process parameter:\n{json.dumps(param, indent=4)}')

            full_name = param["data_source"] + '/' + param["name"]
            if full_name in list_of_parameters:
                value = list_of_parameters[full_name]
                parameter_id = param["id"]
                param = parameters[value]
                data = param['data']
                attrs = param['attributes']

                unit_info = resolve_unit_info(units, attrs)
                attrs |= unit_info
                param = create_parameter_data(value, data, attrs)

                res = client_conn.update_parameter_api(parameter_id, param).json()
                if 'id' in res:
                    # logger.info(f'Updated parameter (id:{parameter_id}) successful:\n{json.dumps(res, indent=4)}')
                    mdc_ops_summary.append({"action": "update", "id": res["id"], "device": res["data_source"], "property": res["name"]})
                    list_of_parameters.pop(full_name)
                else:
                    raise RuntimeError(f'Update parameter (id:{parameter_id}) failed:\n{res}')
            else:
                parameter_id = param["id"]
                res = client_conn.delete_parameter_api(parameter_id)
                logger.debug(f'Delete parameter ({param}):\n{res.status_code}')
                if res.status_code in [200, 204]:
                    mdc_ops_summary.append({"action": "delete", "id": param["id"], "device": param["data_source"], "property": param["name"]})
                else:
                    raise RuntimeError(f'Delete parameter (id:{parameter_id}) failed: {res.json()['info']}')
        except Exception as e:
            logger.warning(e)

    for key, value in list_of_parameters.items():
        param = parameters[value]

        data = param['data']
        attrs = param['attributes']
        unit_info = resolve_unit_info(units, attrs)
        attrs |= unit_info
        param = create_parameter_data(value, data, attrs)

        param['data_groups_parameters_attributes'] = [{'data_group_id': data_group_id}]
        # param['runs_parameters_attributes'] = [{'run_id': 2004}]

        res = client_conn.create_parameter_api(param).json()
        # logger.info(f'Insert parameter successful:\n{res}')
        mdc_ops_summary.append({"action": "create", "id": res["id"], "device": res["data_source"], "property": res["name"]})

    return mdc_ops_summary



import utils.DataView

parameters = [
    "CONTROL/SQS_KBS_CSLIT/MOTOR/SLIT_Y_BOTTOM/encoderPosition",
    "CONTROL/SQS_NQS_CRSC/TSYS/PARKER_TRIGGER/actualControl",
    "CONTROL/SQS_NQS_CSLIT/PMOTOR/RESERVED_4/actualVelocity",
    "CONTROL/SQS_TDK/POWER_SUPPLY/MV1/current/actual",
    "CONTROL/SQS_TDK/POWER_SUPPLY/MV1/voltage/actual",
    "CONTROL/SQS_RACK_MPOD-2/MDL/MPOD_MAPPER/XGMD_MP_gain/channelMeasurementCurrent",
    "CONTROL/SQS_RACK_MPOD-1/MDL/MPOD_MAPPER/strk_ret4_v1/channelSupervisionMaxTerminalVoltage",
    "CONTROL/XFEL_MAGNETS_MAGNETML/DOOCS/KSNEG_2787_T4/pollingInterval",
    "CONTROL/SQS_XTD10_ESLIT/MDL/MAIN/actualPosition",
    "CONTROL/SQS_DIAG1_XGMD/XGM/DOOCS/pulseEnergy/crossUsed",
    "CONTROL/SA3_XTD10_XGM/XGM/DOOCS/current/bottom/rangeCode",
    "CONTROL/SA3_XTD10_VAC/MDL/GATT_P_CELL/weightingCoefficientLR",
    "CONTROL/SQS_KBS_CSLIT/MOTOR/SLIT_Y_TOP/actualPosition",
    "CONTROL/SQS_KBS_CSLIT/MOTOR/SLIT_X_RIGHT/actualPosition",
    "CONTROL/SQS_KBS_CSLIT/MDL/MAIN/actualPositionX",
    "CONTROL/SQS_F3_MOV/MOTOR/Y_US_L/actualVelocity",
    "CONTROL/SQS_RACK_MPOD-2/MDL/MPOD_MAPPER/state/value"
]


def h5_list():
    home = Path.home()
    root_data_path = home / 'gpfs/exfel/data/scs'
    filename = 'RAW-R0007-DA01-S00001.h5'
    file_path = root_data_path / filename

    # Scan all run folders within the data path
    files_list = [f for f in root_data_path.iterdir() if f.suffix == ".h5"]

    print(f"List of files in {root_data_path}:\n{files_list}")

    file_view = H5FileView()
    for file_path in files_list:
        try:
            print(f"List all datasets in the HDF5 file: {file_path}")
            file_view(file_path)
        except Exception as e:
            print(e)

    #  print(json.dumps(file_view, indent=4, cls=H5JsonEncoder))

    for filesname, data_view in file_view.data.items():
        for p in parameters:
            metadata = data_view.get_metadata(p)
            if metadata:
                print(f"{p} ==> {metadata}")


def h5_list_v0():
    home = Path.home()
    root_data_path = home / 'gpfs/exfel/data/scs'
    filename = 'RAW-R0007-DA01-S00001.h5'
    file_path = root_data_path / filename

    # Scan all run folders within the data path
    files_list = [f for f in root_data_path.iterdir() if f.suffix == ".h5"]

    print(f"List of files in {root_data_path}:\n{files_list}")

    h5_index = H5DataIndex('CONTROL')
    for file_path in files_list:
        try:
            print(f"List all datasets in the HDF5 file: {file_path}")
            h5_index.register_file(file_path)
            with h5py.File(file_path, 'r') as f:
                # this will visit all objects inside the hdf5 file and store datasets in h5_index object (ie names)
                f["CONTROL"].visititems(h5_index)
                f.close()
        except Exception as e:
            logger.error(e)

    print(json.dumps(h5_index.data, indent=4))

    paths = []
    for p in parameters:
        paths += [h5_index.get_metadata(p)]

    print(paths)

    compact = {}

    for p in paths:
        f = p["file"]
        compact[f] = compact.get(f, []) + [p["key"]]

    print(json.dumps(compact, indent=4))
